@@ -26,6 +26,8 @@
 
 #include <netinet/in.h>
 
+#include <ndpi_protocol_ids.h>
+
 #include "nd-apps.hpp"
 #include "nd-detection.hpp"
 #include "nd-except.hpp"
@@ -34,20 +36,22 @@
 #include "nd-risks.hpp"
 #include "nd-sha1.h"
 #include "nd-tls-alpn.hpp"
+#include "nd-types.hpp"
 #include "nd-util.hpp"
 
-#include "ndpi_protocol_ids.h"
-
 using namespace std;
 
-// Enable flow hash cache debug logging
+// Enable extra flow hash cache debug logging
 // #define _ND_LOG_FHC 1
 
-// Enable extra flow debug logging
-// #define _ND_LOG_FLOW_DEBUG 1
+// Enable extra DNS flow debug logging
+// #define _ND_LOG_FLOW_DEBUG_DNS 1
 
-// Enable to log STUN debug
-// #define _ND_LOG_STUN 1
+// Enable extra TLS flow debug logging
+// #define _ND_LOG_FLOW_DEBUG_TLS 1
+
+// Enable extra STUN flow debug logging
+// #define _ND_LOG_DEBUG_STUN 1
 
 #define ndEF    entry->flow
 #define ndEFNF  entry->flow->ndpi_flow
@@ -55,18 +59,18 @@ using namespace std;
 
 ndDetectionThread::ndDetectionThread(int16_t cpu, const string &tag,
 #ifdef _ND_ENABLE_NETLINK
-  ndNetlink *netlink,
-#endif
+  ndNetlinkRouteThread *thread_netlink,
 #ifdef _ND_ENABLE_CONNTRACK
-  ndConntrackThread *thread_conntrack,
+  ndNetlinkConntrackThread *thread_conntrack,
+#endif
 #endif
   ndDNSHintCache *dhc, ndFlowHashCache *fhc, uint8_t private_addr)
   : ndThread(tag, (long)cpu, true), ndInstanceClient(),
 #ifdef _ND_ENABLE_NETLINK
-    netlink(netlink),
-#endif
+    thread_netlink(thread_netlink),
 #ifdef _ND_ENABLE_CONNTRACK
     thread_conntrack(thread_conntrack),
+#endif
 #endif
     ndpi(nullptr), dhc(dhc), fhc(fhc) {
     Reload();
@@ -138,7 +142,7 @@ void ndDetectionThread::Reload(void) {
     ndpi = nd_ndpi_init();
 }
 
-void ndDetectionThread::QueuePacket(nd_flow_ptr &flow,
+void ndDetectionThread::QueuePacket(ndFlow::Ptr &flow,
   const ndPacket *packet,
   const uint8_t *data,
   uint16_t length) {
@@ -326,7 +330,7 @@ void ndDetectionThread::ProcessPacket(ndDetectionQueueEntry *entry) {
                     SetDetectedProtocol(entry, result.proto_id);
                 }
 
-                if (result.app_id != ND_APP_UNKNOWN) {
+                if (result.app_id != ndApp::Id::UNKNOWN) {
 #ifdef _ND_LOG_FHC
                     string app_from;
                     ndi.apps.Lookup(
@@ -363,50 +367,22 @@ void ndDetectionThread::ProcessPacket(ndDetectionQueueEntry *entry) {
         memset(ndEFNF, 0, sizeof(ndpi_flow_struct));
     }
 
-    ndpi_protocol ndpi_rc = ndpi_detection_process_packet(ndpi,
+    ndEF->detected_ndpi_protocol = ndpi_detection_process_packet(ndpi,
       ndEFNF, entry->data, entry->length,
       ndEF->ts_last_seen.load(), nullptr);
 
     lock_guard<recursive_mutex> lg(ndEF->lock);
 
-    if (ndpi_rc.master_protocol == NDPI_PROTOCOL_STUN &&
-      ndpi_rc.app_protocol != NDPI_PROTOCOL_UNKNOWN)
-    {
-#ifdef _ND_LOG_STUN
-        nd_dprintf(
-          "%s: STUN[%lu]: master: %hu, app: %hu, stack[0]: "
-          "%hu, stack[1]: %hu, detected_protocol: %hu\n",
-          tag.c_str(), ndEF->stats.detection_packets.load(),
-          ndpi_rc.master_protocol, ndpi_rc.app_protocol,
-          ndEFNF->detected_protocol_stack[0],
-          ndEFNF->detected_protocol_stack[1],
-          ndEF->detected_protocol);
-#endif
-        ndProto::Id id = ProtocolLookup(ndpi_rc.app_protocol, entry);
-
-        if (id != ndProto::Id::TODO) {
-            SetDetectedProtocol(entry, id);
-#ifdef _ND_LOG_STUN
-            nd_dprintf(
-              "%s: STUN[%lu]: refined detected protocol: "
-              "%hu\n",
-              tag.c_str(),
-              ndEF->stats.detection_packets.load(),
-              ndEF->detected_protocol);
-#endif
-        }
-    }
-
     if (ndEF->detected_protocol == ndProto::Id::UNKNOWN) {
-        if (ndpi_rc.master_protocol != NDPI_PROTOCOL_UNKNOWN)
+        if (ndEF->detected_ndpi_protocol.proto.master_protocol != NDPI_PROTOCOL_UNKNOWN)
         {
             SetDetectedProtocol(entry,
-              ProtocolLookup(ndpi_rc.master_protocol, entry));
+              ProtocolLookup(ndEF->detected_ndpi_protocol.proto.master_protocol, entry));
         }
-        else if (ndpi_rc.app_protocol != NDPI_PROTOCOL_UNKNOWN)
+        else if (ndEF->detected_ndpi_protocol.proto.app_protocol != NDPI_PROTOCOL_UNKNOWN)
         {
             SetDetectedProtocol(entry,
-              ProtocolLookup(ndpi_rc.app_protocol, entry));
+              ProtocolLookup(ndEF->detected_ndpi_protocol.proto.app_protocol, entry));
         }
     }
 
@@ -418,13 +394,12 @@ void ndDetectionThread::ProcessPacket(ndDetectionQueueEntry *entry) {
 
     if (ndEF->detected_protocol == ndProto::Id::TODO) {
         char proto_name[64];
-        ndpi_protocol2name(ndpi, ndpi_rc, proto_name,
+        ndpi_protocol2name(ndpi, ndEF->detected_ndpi_protocol, proto_name,
           sizeof(proto_name));
         nd_dprintf(
-          "%s: unmapped protocol detected: ID #%hu/%hu "
-          "(%s)\n",
-          tag.c_str(), ndpi_rc.master_protocol,
-          ndpi_rc.app_protocol, proto_name);
+          "%s: unmapped protocol detected: ID #%hu/%hu (%s)\n", tag.c_str(),
+          ndEF->detected_ndpi_protocol.proto.master_protocol,
+          ndEF->detected_ndpi_protocol.proto.app_protocol, proto_name);
 
         DetectionComplete(entry);
         return;
@@ -433,6 +408,28 @@ void ndDetectionThread::ProcessPacket(ndDetectionQueueEntry *entry) {
     if (ndEF->detected_protocol != ndProto::Id::UNKNOWN)
         DetectionUpdate(entry);
 
+    if (ndEF->GetMasterProtocol() == ndProto::Id::DNS) {
+#ifdef _ND_LOG_FLOW_DEBUG_DNS
+        nd_dprintf(
+          "%s: DNS: extra_packets_func: %p, "
+          "detection_packets: %hhu, "
+          "num_extra_packets_checked: %hhu, "
+          "max_extra_packets_to_check: %hhu: %s\n",
+          tag.c_str(),
+          ndEFNF->extra_packets_func,
+          ndEF->stats.detection_packets.load(),
+          ndEFNF->num_extra_packets_checked,
+          ndEFNF->max_extra_packets_to_check,
+          (ndEFNF->host_server_name[0] != '\0') ?
+            ndEFNF->host_server_name :
+            "(null)");
+#endif
+        //if (ndEFNF->extra_packets_func == nullptr)
+        //    ndEF->Release();
+
+        //return;
+    }
+
     if ((ndEF->flags.detection_init.load() &&
           ndEFNF->extra_packets_func == nullptr) ||
       ndEFNF->num_extra_packets_checked > ndEFNF->max_extra_packets_to_check ||
@@ -489,16 +486,24 @@ void ndDetectionThread::ProcessFlow(ndDetectionQueueEntry *entry) {
     }
 
 #ifdef _ND_ENABLE_CONNTRACK
-    if (thread_conntrack != nullptr &&
-      ndEF->iface->role == ndInterfaceRole::WAN &&
-      ((ndEF->lower_type == ndAddr::Type::LOCAL &&
-         ndEF->upper_type == ndAddr::Type::OTHER) ||
-        (ndEF->lower_type == ndAddr::Type::OTHER &&
-          ndEF->upper_type == ndAddr::Type::LOCAL)))
-    {
-        // Update flow with any collected information from the
-        // connection tracker (CT ID, mark, NAT'd).
+    if (thread_conntrack != nullptr) {
+        // Possibly update flow with metadata collected from the
+        // connection tracker (mark, NAT, counters, etc.).
+#if defined(_ND_ENABLE_CONNTRACK_MDATA)
         thread_conntrack->UpdateFlow(ndEF);
+#else
+        if ((ndEF->iface->role == ndInterfaceRole::WAN &&
+            ((ndEF->lower_type == ndAddr::Type::LOCAL &&
+            ndEF->upper_type == ndAddr::Type::OTHER) ||
+            (ndEF->lower_type == ndAddr::Type::OTHER &&
+            ndEF->upper_type == ndAddr::Type::LOCAL)))
+#if defined(_ND_ENABLE_CONNTRACK_COUNTERS)
+            || ndEF->iface->conntrack_counters
+#endif
+            ) {
+            thread_conntrack->UpdateFlow(ndEF);
+        }
+#endif
     }
 #endif
 
@@ -672,14 +677,14 @@ void ndDetectionThread::SetDetectedProtocol(
 }
 
 void ndDetectionThread::SetDetectedApplication(
-  ndDetectionQueueEntry *entry, nd_app_id_t id) {
-    if (id == ND_APP_UNKNOWN) return;
+  ndDetectionQueueEntry *entry, ndApp::Id id) {
+    if (id == ndApp::Id::UNKNOWN) return;
 
     ndEF->detected_application = id;
     ndi.apps.Lookup(id, ndEF->detected_application_name);
 
     ndEF->category.application = ndi.categories.Lookup(
-      ndCategories::Type::APP, id);
+      ndCategories::Type::APP, static_cast<unsigned>(id));
 }
 
 void ndDetectionThread::DispatchEvent(ndDetectionQueueEntry *entry) {
@@ -692,7 +697,7 @@ void ndDetectionThread::DispatchEvent(ndDetectionQueueEntry *entry) {
 
     ndi.plugins.BroadcastProcessorEvent(event, ndEF);
 
-    if (ndGC_DEBUG || ndGC.h_flow != stderr) {
+    if (ndGC_DEBUG || ndGC.h_debug_log_flow != stderr) {
         bool flow_print = false;
         ndFlags<ndFlow::PrintFlags> flags = ndFlow::PrintFlags::METADATA;
 
@@ -722,7 +727,7 @@ void ndDetectionThread::DispatchEvent(ndDetectionQueueEntry *entry) {
                 }
             }
         }
-        else if (ndGC_VERBOSE || ndGC.h_flow != stderr)
+        else if (ndGC_VERBOSE || ndGC.h_debug_log_flow != stderr)
             flow_print = true;
 
         switch (event) {
@@ -750,10 +755,61 @@ void ndDetectionThread::DispatchEvent(ndDetectionQueueEntry *entry) {
 
 void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
 
-    if (ndEF->flags.detection_init.load() == false) {
+    bool refined = false;
+
+    if (ndEF->detected_ndpi_protocol.proto.master_protocol == NDPI_PROTOCOL_STUN &&
+      ndEF->detected_ndpi_protocol.proto.app_protocol != NDPI_PROTOCOL_UNKNOWN) {
+#ifdef _ND_LOG_DEBUG_STUN
+        nd_dprintf(
+          "%s: STUN[%lu]: master: %hu, app: %hu, stack[0]: "
+          "%hu, stack[1]: %hu, detected_protocol: %hu\n",
+          tag.c_str(), ndEF->stats.detection_packets.load(),
+          ndEF->detected_ndpi_protocol.proto.master_protocol,
+          ndEF->detected_ndpi_protocol.proto.app_protocol,
+          ndEFNF->detected_protocol_stack[0],
+          ndEFNF->detected_protocol_stack[1],
+          ndEF->detected_protocol);
+#endif
+        ndProto::Id id = ProtocolLookup(
+          ndEF->detected_ndpi_protocol.proto.app_protocol, entry);
+
+        if (id != ndProto::Id::TODO && id != ndEF->detected_protocol) {
+            refined = true;
+            SetDetectedProtocol(entry, id);
+#ifdef _ND_LOG_DEBUG_STUN
+            nd_dprintf(
+              "%s: STUN[%lu]: refined detected protocol: %hu\n",
+              tag.c_str(),
+              ndEF->stats.detection_packets.load(),
+              ndEF->detected_protocol);
+#endif
+        }
+    }
+
+    ndApp::Id app_id = ndApp::Id::UNKNOWN;
+
+    if (ndEF->lower_type == ndAddr::Type::OTHER) {
+        app_id = ndi.apps.Find(ndEF->lower_addr, true);
+
+        if (ndEF->detected_application == ndApp::Id::UNKNOWN)
+            app_id = ndi.apps.Find(ndEF->upper_addr, true);
+    }
+    else {
+        app_id = ndi.apps.Find(ndEF->upper_addr, true);
+
+        if (ndEF->detected_application == ndApp::Id::UNKNOWN)
+            app_id = ndi.apps.Find(ndEF->lower_addr, true);
+    }
+
+    if (app_id != ndApp::Id::UNKNOWN) {
+        ndEF->flags.app_ip_override = true;
+        SetDetectedApplication(entry, app_id);
+        ndEF->Print(ndFlow::PrintFlags::METADATA, "aio");
+    }
+    else if (refined || ndEF->flags.detection_init.load() == false) {
         // Determine application based on master protocol metadata
         // for Protocol / Application "Twins"
-        nd_app_id_t app_twin = ND_APP_UNKNOWN;
+        ndApp::Id app_twin = ndApp::Id::UNKNOWN;
 
         switch (ndEF->detected_protocol) {
         case ndProto::Id::APPLE_PUSH:
@@ -764,10 +820,12 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
             app_twin = ndi.apps.Lookup("netify.avast");
             break;
 
+        case ndProto::Id::DINGTALK:
+            app_twin = ndi.apps.Lookup("netify.dingtalk");
+            break;
+
         case ndProto::Id::FACEBOOK_VOIP:
-            // TODO: netify.meta-messaging or netify.facebook-messager?
-            app_twin = ndi.apps.Lookup(
-              "netify.meta-messaging");
+            app_twin = ndi.apps.Lookup("netify.meta");
             break;
 
         case ndProto::Id::LINE_CALL:
@@ -778,6 +836,18 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
             app_twin = ndi.apps.Lookup("netify.nest");
             break;
 
+        case ndProto::Id::NTP:
+            app_twin = ndi.apps.Lookup("netify.ntp");
+            break;
+
+        case ndProto::Id::OOKLA:
+            app_twin = ndi.apps.Lookup("netify.speedtest");
+            break;
+
+        case ndProto::Id::QQ:
+            app_twin = ndi.apps.Lookup("netify.qq");
+            break;
+
         case ndProto::Id::SPOTIFY:
             app_twin = ndi.apps.Lookup("netify.spotify");
             break;
@@ -791,6 +861,10 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
             app_twin = ndi.apps.Lookup("netify.teams");
             break;
 
+        case ndProto::Id::SNAPCHAT_CALL:
+            app_twin = ndi.apps.Lookup("netify.snapchat");
+            break;
+
         case ndProto::Id::STEAM:
             app_twin = ndi.apps.Lookup("netify.steam");
             break;
@@ -834,7 +908,7 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
         default: break;
         }
 
-        if (app_twin != ND_APP_UNKNOWN)
+        if (app_twin != ndApp::Id::UNKNOWN)
             SetDetectedApplication(entry, app_twin);
     }
 
@@ -842,6 +916,34 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
         bool updated = false;
 
         switch (ndEF->GetMasterProtocol()) {
+        case ndProto::Id::DNS:
+            if (ndEFNF->host_server_name[0] != '\0') {
+                if (ndEF->host_server_name.empty()) {
+                    updated = true;
+                    SetHostServerName(entry, ndEFNF->host_server_name);
+                }
+                else {
+                    string hostname;
+                    hostname.assign(ndEFNF->host_server_name);
+                    nd_set_hostname(hostname);
+                    if (hostname != ndEF->host_server_name) {
+                        updated = true;
+                        ndEF->host_server_name = hostname;
+                    }
+                }
+
+                if (updated &&
+                    ndEF->detected_application == ndApp::Id::UNKNOWN) {
+                    SetDetectedApplication(entry,
+                      ndi.apps.Find(ndEF->tls.server_cn));
+                }
+#ifdef _ND_LOG_FLOW_DEBUG_DNS
+                nd_dprintf("%s: DNS: %s hostname.\n", tag.c_str(),
+                    (updated) ? "updated" : "set");
+#endif
+            }
+            break;
+
         case ndProto::Id::TLS:
         case ndProto::Id::DTLS:
         case ndProto::Id::QUIC:
@@ -850,7 +952,7 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
             {
                 updated = true;
                 ndEF->tls.version = ndEFNFP.tls_quic.ssl_version;
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf("%s: TLS: set version.\n", tag.c_str());
 #endif
             }
@@ -860,20 +962,20 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
             {
                 updated = true;
                 ndEF->tls.cipher_suite = ndEFNFP.tls_quic.server_cipher;
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf("%s: TLS: set cipher_suite.\n",
                   tag.c_str());
 #endif
             }
 
-            if (ndEF->tls.server_ja3.empty() &&
-              ndEFNFP.tls_quic.ja3_server[0] != '\0')
+            if (ndEF->tls.client_ja4.empty() &&
+              ndEFNFP.tls_quic.ja4_client[0] != '\0')
             {
                 updated = true;
-                ndEF->tls.server_ja3.assign(
-                  ndEFNFP.tls_quic.ja3_server);
-#ifdef _ND_LOG_FLOW_DEBUG
-                nd_dprintf("%s: TLS: set ja3_server.\n",
+                ndEF->tls.client_ja4.assign(
+                  ndEFNFP.tls_quic.ja4_client);
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
+                nd_dprintf("%s: TLS: set ja4_client.\n",
                   tag.c_str());
 #endif
             }
@@ -885,12 +987,11 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
                 nd_set_hostname(ndEF->tls.server_cn,
                   ndEFNFP.tls_quic.serverCN,
                   ND_FLOW_TLS_CNLEN);
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf("%s: TLS: set serverCN: %s\n",
                   tag.c_str(), ndEF->tls.server_cn.c_str());
 #endif
-                if (ndEF->detected_application == ND_APP_UNKNOWN)
-                {
+                if (ndEF->detected_application == ndApp::Id::UNKNOWN) {
                     SetDetectedApplication(entry,
                       ndi.apps.Find(ndEF->tls.server_cn));
                 }
@@ -902,7 +1003,7 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
                 updated = true;
                 ndEF->tls.issuer_dn.assign(ndEFNFP.tls_quic.issuerDN);
                 nd_trim(ndEF->tls.issuer_dn);
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf("%s: TLS: set issuerDN.\n", tag.c_str());
 #endif
             }
@@ -913,7 +1014,7 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
                 updated = true;
                 ndEF->tls.subject_dn.assign(ndEFNFP.tls_quic.subjectDN);
                 nd_trim(ndEF->tls.issuer_dn);
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf("%s: TLS: set subjectDN.\n",
                   tag.c_str());
 #endif
@@ -928,7 +1029,7 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
                 memcpy(&ndEF->tls.cert_fingerprint[0],
                   ndEFNFP.tls_quic.sha1_certificate_fingerprint,
                   SHA1_DIGEST_LENGTH);
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf(
                   "%s: TLS: set cert_fingerprint.\n", tag.c_str());
 #endif
@@ -937,7 +1038,7 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
             if (ndEF->tls.alpn.empty() && ndEFNFP.tls_quic.advertised_alpns)
             {
                 updated = ProcessALPN(entry, true);
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf("%s: TLS: set ALPNs.\n", tag.c_str());
 #endif
             }
@@ -946,7 +1047,7 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
               ndEFNFP.tls_quic.negotiated_alpn)
             {
                 updated = ProcessALPN(entry, false);
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf(
                   "%s: TLS: set negotiated_alpn.\n", tag.c_str());
 #endif
@@ -958,7 +1059,7 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
                 updated = true;
                 ndEF->tls.ech.version =
                   ndEFNFP.tls_quic.encrypted_ch.version;
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf("%s: TLS: set ECH.\n", tag.c_str());
 #endif
             }
@@ -971,16 +1072,16 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
                   ndEFNFP.tls_quic.encrypted_sni.cipher_suite;
                 ndEF->tls.esni.esni.assign(
                   ndEFNFP.tls_quic.encrypted_sni.esni);
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf("%s: TLS: set ESNI.\n", tag.c_str());
 #endif
             }
 
-            if (ndEFNFP.tls_quic.hello_processed &&
+            if (ndEFNFP.tls_quic.client_hello_processed &&
               ! ndEF->tls.proc_hello.load())
             {
                 ndEF->tls.proc_hello = true;
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf("%s: TLS: hello processed.\n",
                   tag.c_str());
 #endif
@@ -990,7 +1091,7 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
               ! ndEF->tls.proc_certificate.load())
             {
                 ndEF->tls.proc_certificate = true;
-#ifdef _ND_LOG_FLOW_DEBUG
+#ifdef _ND_LOG_FLOW_DEBUG_TLS
                 nd_dprintf(
                   "%s: TLS: certificate processed.\n", tag.c_str());
 #endif
@@ -1060,6 +1161,127 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
         default: break;
         }
 
+        if (ndEF->detected_ndpi_protocol.proto.master_protocol == NDPI_PROTOCOL_STUN) {
+#ifdef _ND_LOG_DEBUG_STUN
+            nd_dprintf("%s: STUN: maybe_dtls: %s, rtcp_seen: %s, "
+                "is_turn: %s, is_client_controlling: %s\n",
+                tag.c_str(),
+                (ndEFNF->stun.maybe_dtls) ? "yes" : "no",
+                (ndEFNF->stun.rtcp_seen) ? "yes" : "no",
+                (ndEFNF->stun.is_turn) ? "yes" : "no",
+                (ndEFNF->stun.is_client_controlling) ? "yes" : "no"
+            );
+#endif
+            enum class STUN_AddrType : uint8_t {
+                MAPPED, PEER, RELAYED, RESPONSE, OTHER
+            };
+
+#ifdef _ND_LOG_DEBUG_STUN
+            auto STUN_Addr = [this, entry] (
+#else
+            auto STUN_Addr = [entry] (
+#endif
+                STUN_AddrType type, const ndpi_address_port &addr) {
+
+                ndAddr *_addr = nullptr;
+
+                switch (type) {
+                case STUN_AddrType::MAPPED:
+                    _addr = &ndEF->stun.mapped;
+                    break;
+                case STUN_AddrType::PEER:
+                    _addr = &ndEF->stun.peer;
+                    break;
+                case STUN_AddrType::RELAYED:
+                    _addr = &ndEF->stun.relayed;
+                    break;
+                case STUN_AddrType::RESPONSE:
+                    _addr = &ndEF->stun.response;
+                    break;
+                case STUN_AddrType::OTHER:
+                    _addr = &ndEF->stun.other;
+                    break;
+                }
+
+                if (addr.is_ipv6) {
+                    bool valid = false;
+                    struct sockaddr_in6 sin6 = {
+                        .sin6_family = AF_INET6
+                    };
+
+                    for (auto i = 0; i < 8; i++) {
+                        if (addr.address.v6[i]) valid = true;
+                        sin6.sin6_addr.s6_addr[i] = addr.address.v6[i];
+                    }
+
+                    if (! valid) return;
+
+                    ndAddr::Create(*_addr, &sin6);
+                }
+                else if (addr.address.v4) {
+                    struct sockaddr_in sin = {
+                        .sin_family = AF_INET
+                    };
+
+                    sin.sin_addr.s_addr = addr.address.v4;
+
+                    ndAddr::Create(*_addr, &sin);
+                }
+                else return;
+
+                _addr->SetPort(htons(addr.port));
+
+#ifdef _ND_LOG_DEBUG_STUN
+                string address;
+                ndAddr::MakeString(*_addr, address);
+
+                switch (type) {
+                case STUN_AddrType::MAPPED:
+                    nd_printf("%s: STUN: mapped: %s\n",
+                        tag.c_str(), address.c_str());
+                    break;
+                case STUN_AddrType::PEER:
+                    nd_printf("%s: STUN: peer: %s\n",
+                        tag.c_str(), address.c_str());
+                    break;
+                case STUN_AddrType::RELAYED:
+                    nd_printf("%s: STUN: relayed: %s\n",
+                        tag.c_str(), address.c_str());
+                    break;
+                case STUN_AddrType::RESPONSE:
+                    nd_printf("%s: STUN: response: %s\n",
+                        tag.c_str(), address.c_str());
+                    break;
+                case STUN_AddrType::OTHER:
+                    nd_printf("%s: STUN: other: %s\n",
+                        tag.c_str(), address.c_str());
+                    break;
+                }
+#endif
+            };
+
+            if (! ndEF->stun.mapped.IsValid()) {
+                STUN_Addr(STUN_AddrType::MAPPED, ndEFNF->stun.mapped_address);
+                if (! updated) updated = ndEF->stun.mapped.IsValid();
+            }
+            if (! ndEF->stun.peer.IsValid()) {
+                STUN_Addr(STUN_AddrType::PEER, ndEFNF->stun.peer_address);
+                if (! updated) updated = ndEF->stun.peer.IsValid();
+            }
+            if (! ndEF->stun.relayed.IsValid()) {
+                STUN_Addr(STUN_AddrType::RELAYED, ndEFNF->stun.relayed_address);
+                if (! updated) updated = ndEF->stun.relayed.IsValid();
+            }
+            if (! ndEF->stun.response.IsValid()) {
+                STUN_Addr(STUN_AddrType::RESPONSE, ndEFNF->stun.response_origin);
+                if (! updated) updated = ndEF->stun.response.IsValid();
+            }
+            if (! ndEF->stun.other.IsValid()) {
+                STUN_Addr(STUN_AddrType::OTHER, ndEFNF->stun.other_address);
+                if (! updated) updated = ndEF->stun.other.IsValid();
+            }
+        }
+
         if (ndEF->flags.detection_init.load()) {
             if (updated) {
                 ndEF->flags.detection_updated = true;
@@ -1080,15 +1302,16 @@ void ndDetectionThread::DetectionUpdate(ndDetectionQueueEntry *entry) {
 
 void ndDetectionThread::DetectionGuess(ndDetectionQueueEntry *entry) {
     uint8_t guessed = 0;
-    ndpi_protocol ndpi_rc = ndpi_detection_giveup(ndpi,
-      ndEFNF, 1, &guessed);
+    ndEF->detected_ndpi_protocol = ndpi_detection_giveup(ndpi,
+      ndEFNF, &guessed);
 
     if (guessed) {
         ndProto::Id id = ProtocolLookup(
-          ndpi_rc.master_protocol, entry);
+          ndEF->detected_ndpi_protocol.proto.master_protocol, entry);
 
         if (id == ndProto::Id::UNKNOWN) {
-            id = ProtocolLookup(ndpi_rc.app_protocol, entry);
+            id = ProtocolLookup(
+              ndEF->detected_ndpi_protocol.proto.app_protocol, entry);
             if (id != ndProto::Id::UNKNOWN) {
                 SetDetectedProtocol(entry, id);
             }
@@ -1107,7 +1330,7 @@ void ndDetectionThread::DetectionComplete(
 
     // Determine application by DHC (dns_host_name) if still
     // unknown.
-    if (ndEF->detected_application == ND_APP_UNKNOWN) {
+    if (ndEF->detected_application == ndApp::Id::UNKNOWN) {
         if (! ndEF->dns_host_name.empty()) {
             if (! ndGC_DHC_PARTIAL_LOOKUPS ||
               ndEF->host_server_name.empty())
@@ -1115,10 +1338,10 @@ void ndDetectionThread::DetectionComplete(
                 SetDetectedApplication(entry,
                   ndi.apps.Find(ndEF->dns_host_name));
 
-                if (ndEF->detected_application == ND_APP_UNKNOWN)
+                if (ndEF->detected_application == ndApp::Id::UNKNOWN)
                 {
                     ndEF->flags.dhc_hit = (ndEF->detected_application !=
-                      ND_APP_UNKNOWN);
+                      ndApp::Id::UNKNOWN);
                 }
             }
         }
@@ -1127,13 +1350,13 @@ void ndDetectionThread::DetectionComplete(
     // Determine application by network CIDR if still unknown.
     // DNS flows excluded...
     if (ndEF->GetMasterProtocol() != ndProto::Id::DNS &&
-      ndEF->detected_application == ND_APP_UNKNOWN)
+      ndEF->detected_application == ndApp::Id::UNKNOWN)
     {
         if (ndEF->lower_type == ndAddr::Type::OTHER) {
             SetDetectedApplication(entry,
               ndi.apps.Find(ndEF->lower_addr));
 
-            if (ndEF->detected_application == ND_APP_UNKNOWN)
+            if (ndEF->detected_application == ndApp::Id::UNKNOWN)
                 SetDetectedApplication(entry,
                   ndi.apps.Find(ndEF->upper_addr));
         }
@@ -1141,7 +1364,7 @@ void ndDetectionThread::DetectionComplete(
             SetDetectedApplication(entry,
               ndi.apps.Find(ndEF->upper_addr));
 
-            if (ndEF->detected_application == ND_APP_UNKNOWN)
+            if (ndEF->detected_application == ndApp::Id::UNKNOWN)
                 SetDetectedApplication(entry,
                   ndi.apps.Find(ndEF->lower_addr));
         }
@@ -1154,19 +1377,22 @@ void ndDetectionThread::DetectionComplete(
             ndEF->flags.soft_dissector = true;
 
             if (nsd.aid > -1) {
-                if (nsd.aid == ND_APP_UNKNOWN) {
-                    ndEF->detected_application = ND_APP_UNKNOWN;
+                if (nsd.aid == static_cast<int>(ndApp::Id::UNKNOWN)) {
+                    ndEF->detected_application = ndApp::Id::UNKNOWN;
                     ndEF->detected_application_name.clear();
                     ndEF->category.application = ND_CAT_UNKNOWN;
                 }
                 else {
                     SetDetectedApplication(entry,
-                      (nd_app_id_t)nsd.aid);
+                      static_cast<ndApp::Id>(nsd.aid));
                 }
             }
 
-            if (nsd.pid > -1)
-                SetDetectedProtocol(entry, (ndProto::Id)nsd.pid);
+            if (nsd.pid > -1) {
+                SetDetectedProtocol(
+                    entry, static_cast<ndProto::Id>(nsd.pid)
+                );
+            }
         }
     }
 
@@ -1175,7 +1401,7 @@ void ndDetectionThread::DetectionComplete(
     if (fhc != nullptr && ! ndEF->flags.fhc_hit.load() &&
       ! ndEF->flags.detection_guessed.load() &&
       (ndEF->detected_protocol != ndProto::Id::UNKNOWN ||
-        ndEF->detected_application != ND_APP_UNKNOWN) &&
+        ndEF->detected_application != ndApp::Id::UNKNOWN) &&
       ndEF->ip_protocol != IPPROTO_ICMP &&
       ndEF->ip_protocol != IPPROTO_ICMPV6 &&
       ndEF->ip_protocol != IPPROTO_IGMP &&
