@@ -22,9 +22,11 @@
 #include "config.h"
 #endif
 
+#include <algorithm>
 #include <iomanip>
 #include <locale>
 #include <mutex>
+#include <sstream>
 
 #include "nd-apps.hpp"
 #include "nd-category.hpp"
@@ -63,9 +65,8 @@ void ndFlowStats::UpdateRate(bool lower, uint64_t timestamp,
     rate = (divisor > 0) ? ((float)total / (float)divisor) : 0.0f;
 }
 
-ndFlow::ndFlow(nd_iface_ptr &iface)
-  : iface(iface),
-    detected_protocol_name("Unknown") {
+ndFlow::ndFlow(ndInterface::Ptr &iface)
+  : iface(iface), detected_protocol_name("Unknown") {
 }
 
 ndFlow::ndFlow(const ndFlow &flow)
@@ -77,8 +78,7 @@ ndFlow::ndFlow(const ndFlow &flow)
     ts_last_seen(flow.ts_last_seen.load()),
     tunnel_type(flow.tunnel_type), origin(flow.origin),
     digest_lower(flow.digest_lower), flags{}, tcp{},
-#if defined(_ND_ENABLE_CONNTRACK) && \
-  defined(_ND_ENABLE_CONNTRACK_MDATA)
+#ifdef _ND_ENABLE_CONNTRACK
     conntrack(flow.conntrack),
 #endif
     gtp(flow.gtp), detected_protocol_name("Unknown") {
@@ -149,7 +149,7 @@ void ndFlow::Hash(const string &device, bool hash_mdata,
         sha1_write(&ctx, (const char *)&detected_protocol,
           sizeof(ndProto::Id));
         sha1_write(&ctx, (const char *)&detected_application,
-          sizeof(nd_app_id_t));
+          sizeof(ndApp::Id));
 
         if (! host_server_name.empty()) {
             sha1_write(&ctx, host_server_name.c_str(),
@@ -170,11 +170,7 @@ void ndFlow::Hash(const string &device, bool hash_mdata,
     else {
         ndDigest mdata;
         sha1_result(&ctx, mdata.data());
-
-        if (! digest_mdata.empty() && mdata != digest_mdata.back())
-        {
-            digest_mdata.push_back(mdata);
-        }
+        PushMetadataDigest(mdata);
     }
 }
 
@@ -199,12 +195,11 @@ ndProto::Id ndFlow::GetMasterProtocol(void) const {
     case ndProto::Id::MQTTS:
     case ndProto::Id::NNTPS:
     case ndProto::Id::SIPS: return ndProto::Id::TLS;
+    case ndProto::Id::HLS:
     case ndProto::Id::HTTP:
     case ndProto::Id::HTTP_CONNECT:
     case ndProto::Id::HTTP_PROXY:
     case ndProto::Id::OOKLA:
-    case ndProto::Id::PPSTREAM:
-    case ndProto::Id::QQ:
     case ndProto::Id::RTSP:
     case ndProto::Id::STEAM:
     case ndProto::Id::TEAMVIEWER:
@@ -290,16 +285,10 @@ bool ndFlow::HasTLSSubjectDN(void) const {
       ! tls.subject_dn.empty());
 }
 
-bool ndFlow::HasTLSClientJA3(void) const {
+bool ndFlow::HasTLSClientJA4(void) const {
     return ((GetMasterProtocol() == ndProto::Id::TLS ||
               detected_protocol == ndProto::Id::DTLS) &&
-      ! tls.client_ja3.empty());
-}
-
-bool ndFlow::HasTLSServerJA3(void) const {
-    return ((GetMasterProtocol() == ndProto::Id::TLS ||
-              detected_protocol == ndProto::Id::DTLS) &&
-      ! tls.server_ja3.empty());
+      ! tls.client_ja4.empty());
 }
 
 bool ndFlow::HasBTInfoHash(void) const {
@@ -317,6 +306,11 @@ bool ndFlow::HasMDNSDomainName(void) const {
       ! mdns.domain_name.empty());
 }
 
+bool ndFlow::HasSTUNAddress(void) const {
+    return (stun.mapped.IsValid() || stun.peer.IsValid() ||
+      stun.relayed.IsValid() || stun.response.IsValid() || stun.other.IsValid());
+}
+
 void ndFlow::Print(ndFlags<PrintFlags> pflags,
   const string &prefix) const {
     bool multiline = false;
@@ -324,13 +318,15 @@ void ndFlow::Print(ndFlags<PrintFlags> pflags,
 
     lock_guard<recursive_mutex> lg(lock);
 
-    const string &p = (prefix.empty()) ? iface->ifname : prefix;
-    size_t plen = p.size();
+    ostringstream os;
+    if (! prefix.empty()) os << prefix << ": ";
+    os << iface->ifname;
+    size_t plen = 4;
 
     nd_output_lock();
 
     try {
-        dls << p << ": ";
+        dls << os.str() << ": ";
 
         if (ndFlagBoolean(pflags, PrintFlags::HASHES)) {
             for (unsigned i = 0; i < 5; i++) {
@@ -376,6 +372,7 @@ void ndFlow::Print(ndFlags<PrintFlags> pflags,
           << (flags.ip_nat.load() ? 'n' : '-')
           << (risk.risks.empty() ? '-' : 'r')
           << (flags.soft_dissector.load() ? 's' : '-')
+          << (flags.app_ip_override.load() ? 'o' : '-')
           << (tcp.fin_ack.load() ? 'F' : '-')
           << (ndFlagBoolean(privacy_mask,
                 (PrivacyMask::LOWER_MAC | PrivacyMask::LOWER_IP)) ?
@@ -455,7 +452,49 @@ void ndFlow::Print(ndFlags<PrintFlags> pflags,
               << ((! detected_application_name.empty()) ?
                      detected_application_name :
                      "");
+#ifdef _ND_ENABLE_CONNTRACK
+            dls << endl
+                << setw(plen) << " " << ":"
+                << " CT ID: " << conntrack.id;
+#ifdef _ND_ENABLE_CONNTRACK_MDATA
+            dls << ":"
+                << " MARK: " << conntrack.mark;
 
+            if (vlan_id > 0) {
+                dls
+                  << endl
+                  << setw(plen) << " "
+                  << ": VLAN: " << vlan_id;
+            }
+
+            if (conntrack.reply_src_addr.IsValid() ||
+                conntrack.reply_dst_addr.IsValid()) {
+                dls << endl
+                    << setw(plen) << " " << ": CT REPLY";
+                if (conntrack.reply_src_addr.IsValid()) {
+                    dls << " SRC: "
+                        << conntrack.reply_src_addr.GetString() << ":"
+                        << conntrack.reply_src_addr.GetPort();
+                }
+                if (conntrack.reply_dst_addr.IsValid()) {
+                    dls << " DST: "
+                        << conntrack.reply_dst_addr.GetString() << ":"
+                        << conntrack.reply_dst_addr.GetPort();
+                }
+            }
+#endif // _ND_ENABLE_CONNTRACK_MDATA
+#endif // _ND_ENABLE_CONNTRACK
+#ifdef _ND_ENABLE_NFQUEUE
+            if (! nfq.src_iface.empty() || ! nfq.dst_iface.empty()) {
+                dls << endl
+                    << setw(plen) << " "
+                    << ": NFQ";
+                if (! nfq.src_iface.empty())
+                    dls << " SRC: " << nfq.src_iface;
+                if (! nfq.dst_iface.empty())
+                    dls << " DST: " << nfq.dst_iface;
+            }
+#endif
             if (! dns_host_name.empty() ||
               ! host_server_name.empty())
             {
@@ -520,6 +559,44 @@ void ndFlow::Print(ndFlags<PrintFlags> pflags,
                     dls << " SSH/SA: " << ssh.server_agent;
             }
 
+            if (HasSTUNAddress()) {
+                if (stun.mapped.IsValid()) {
+                    dls << endl
+                        << setw(plen) << " "
+                        << ":";
+                    dls << " STUN/MAPPED: " << stun.mapped.GetString()
+                        << ":" << stun.mapped.GetPort();
+                }
+                if (stun.peer.IsValid()) {
+                    dls << endl
+                        << setw(plen) << " "
+                        << ":";
+                    dls << " STUN/PEER: " << stun.peer.GetString()
+                        << ":" << stun.peer.GetPort();
+                }
+                if (stun.relayed.IsValid()) {
+                    dls << endl
+                        << setw(plen) << " "
+                        << ":";
+                    dls << " STUN/RELAYED: " << stun.relayed.GetString()
+                        << ":" << stun.relayed.GetPort();
+                }
+                if (stun.response.IsValid()) {
+                    dls << endl
+                        << setw(plen) << " "
+                        << ":";
+                    dls << " STUN/RESPONSE: " << stun.response.GetString()
+                        << ":" << stun.response.GetPort();
+                }
+                if (stun.other.IsValid()) {
+                    dls << endl
+                        << setw(plen) << " "
+                        << ":";
+                    dls << " STUN/OTHER: " << stun.other.GetString()
+                        << ":" << stun.other.GetPort();
+                }
+            }
+
             if ((GetMasterProtocol() == ndProto::Id::TLS ||
                   detected_protocol == ndProto::Id::DTLS ||
                   detected_protocol == ndProto::Id::QUIC) &&
@@ -552,14 +629,17 @@ void ndFlow::Print(ndFlags<PrintFlags> pflags,
             if (HasTLSEncryptedCH()) {
                 dls << endl
                     << setw(plen) << " "
-                    << ": TLS/ECH: v" << hex
-                    << tls.ech.version << dec;
+                    << ": TLS/ECH: 0x" << setfill('0')
+                    << setw(4) << hex << tls.ech.version
+                    << setfill(' ') << dec;
             }
 
             if (HasTLSEncryptedSNI()) {
                 dls << endl
                     << setw(plen) << " "
-                    << ": TLS/ESNI: v" << hex << tls.esni.cipher_suite
+                    << ": TLS/ESNI: 0x" << setfill('0')
+                    << setw(4) << hex
+                    << tls.esni.cipher_suite << setfill(' ')
                     << dec << ": " << tls.esni.esni;
             }
 
@@ -572,6 +652,12 @@ void ndFlow::Print(ndFlags<PrintFlags> pflags,
                 if (HasTLSSubjectDN())
                     dls << " TLS/SDN: " << tls.subject_dn;
             }
+
+            if (HasTLSClientJA4()) {
+                dls << endl
+                    << setw(plen) << " "
+                    << ": JA4: " << tls.client_ja4;
+            }
         }
 
         if (ndFlagBoolean(pflags, PrintFlags::RISKS)) {
@@ -760,3 +846,12 @@ void ndFlow::GetLowerMap(ndAddr::Type lt, ndAddr::Type ut,
     }
 #endif
 }
+
+bool ndFlow::PushMetadataDigest(const ndDigest &digest) {
+    for (auto &i : digest_mdata) {
+        if (equal(i.begin(), i.end(), digest.begin())) return false;
+    }
+
+    digest_mdata.push_back(digest);
+    return true;
+}
