@@ -35,12 +35,20 @@
 #include <pwd.h>
 #include <sys/file.h>
 #include <sys/ioctl.h>
+#include <sys/resource.h>
 #include <sys/socket.h>
 #include <sys/types.h>
 #if defined(__FreeBSD__)
 #include <sys/sysctl.h>
 #include <sys/user.h>
 #endif
+#if defined(__linux__)
+#include <linux/wireless.h>
+#ifdef _ND_ENABLE_LIBCAP
+#include <sys/capability.h>
+#include <sys/prctl.h>
+#endif // _ND_ENABLE_LIBCAP
+#endif
 #include <syslog.h>
 #include <unistd.h>
 #include <zlib.h>
@@ -48,19 +56,24 @@
 #include <array>
 #include <csignal>
 #include <cstdarg>
+#include <cstdio>
 #include <cstdlib>
 #include <cstring>
 #include <deque>
 #include <fstream>
 #include <iomanip>
 #include <iostream>
+#include <memory>
 #include <mutex>
 #include <regex>
+#include <sched.h>
 #include <sstream>
 #include <stdexcept>
 #include <unordered_map>
 #include <vector>
 
+#include "timezone.hpp"
+
 #include "nd-config.hpp"
 #include "nd-except.hpp"
 #include "nd-util.hpp"
@@ -69,7 +82,7 @@
 using namespace std;
 using json = nlohmann::json;
 
-static mutex nd_printf_mutex;
+std::mutex nd_printf_mutex;
 
 const bool nd_isatty = (isatty(STDOUT_FILENO) == 1 &&
   isatty(STDERR_FILENO) == 1);
@@ -88,8 +101,7 @@ const char *ndTerm::Attr::CURSOR_SHOW =
   (nd_isatty) ? "\033[?25h" : "";
 const char *ndTerm::Attr::CLEAR_EOL =
   (nd_isatty) ? "\033[0K" : "";
-const char *ndTerm::Attr::BOLD =
-  (nd_isatty) ? "\033[1m" : "";
+const char *ndTerm::Attr::BOLD = (nd_isatty) ? "\033[1m" : "";
 const char *ndTerm::Attr::UNDERLINE =
   (nd_isatty) ? "\033[4m" : "";
 
@@ -104,14 +116,6 @@ bool ndTerm::IsTTY(void) {
     return nd_isatty;
 }
 
-void *nd_mem_alloc(size_t size) {
-    return malloc(size);
-}
-
-void nd_mem_free(void *ptr) {
-    free(ptr);
-}
-
 void nd_output_lock(void) {
     nd_printf_mutex.lock();
 }
@@ -121,7 +125,7 @@ void nd_output_unlock(void) {
 }
 
 void nd_printf(const char *format, ...) {
-    if (ndGC_QUIET) return;
+    if (ndGC_QUIET && ndGC.h_debug_log == stderr) return;
 
     va_list ap;
     va_start(ap, format);
@@ -130,15 +134,20 @@ void nd_printf(const char *format, ...) {
 }
 
 void nd_printf(const char *format, va_list ap) {
-    if (ndGC_QUIET) return;
+    if (ndGC_QUIET && ndGC.h_debug_log == stderr) return;
 
     lock_guard<mutex> lock(nd_printf_mutex);
 
     vsyslog(LOG_DAEMON | LOG_INFO, format, ap);
+
+    if (ndGC.h_debug_log != stderr) {
+        vfprintf(ndGC.h_debug_log, format, ap);
+        fflush(ndGC.h_debug_log);
+    }
 }
 
 void nd_dprintf(const char *format, ...) {
-    if (! ndGC_DEBUG) return;
+    if (! ndGC_DEBUG && ndGC.h_debug_log == stderr) return;
 
     va_list ap;
     va_start(ap, format);
@@ -147,11 +156,13 @@ void nd_dprintf(const char *format, ...) {
 }
 
 void nd_dprintf(const char *format, va_list ap) {
-    if (! ndGC_DEBUG) return;
+    if (! ndGC_DEBUG && ndGC.h_debug_log == stderr) return;
 
     lock_guard<mutex> lock(nd_printf_mutex);
 
-    vfprintf(stderr, format, ap);
+    vfprintf(ndGC.h_debug_log, format, ap);
+
+    if (ndGC.h_debug_log != stderr) fflush(ndGC.h_debug_log);
 }
 
 void nd_flow_printf(const char *format, ...) {
@@ -159,36 +170,12 @@ void nd_flow_printf(const char *format, ...) {
 
     va_list ap;
     va_start(ap, format);
-    vfprintf(ndGC.h_flow, format, ap);
+    vfprintf(ndGC.h_debug_log_flow, format, ap);
     va_end(ap);
+
+    fflush(ndGC.h_debug_log_flow);
 }
 
-#ifdef _ND_ENABLE_NDPI_DEBUG
-void nd_ndpi_debug_printf(uint32_t protocol, void *ndpi,
-  ndpi_log_level_t level, const char *file,
-  const char *func, unsigned line, const char *format, ...) {
-    if (ndGC_DEBUG && (ndGC_DEBUG_NDPI || level == NDPI_LOG_ERROR))
-    {
-        lock_guard<mutex> lock(nd_printf_mutex);
-
-        va_list ap;
-        va_start(ap, format);
-
-        fprintf(stdout,
-          "[nDPI:%08x:%p:%s]: %s/%s:%d: ", protocol, ndpi,
-          (level == NDPI_LOG_ERROR)         ? "ERROR" :
-            (level == NDPI_LOG_TRACE)       ? "TRACE" :
-            (level == NDPI_LOG_DEBUG)       ? "DEBUG" :
-            (level == NDPI_LOG_DEBUG_EXTRA) ? "DEXTRA" :
-                                              "UNK???",
-          file, func, line);
-
-        vfprintf(stdout, format, ap);
-        va_end(ap);
-    }
-}
-#endif  // NDPI_ENABLE_DEBUG_MESSAGES
-
 int ndLogBuffer::overflow(int ch) {
     if (ch != EOF) os << (char)ch;
 
@@ -217,8 +204,10 @@ int ndDebugLogBuffer::sync() {
 
 int ndDebugLogBufferUnlocked::sync() {
     if (! os.str().empty()) {
-        if (ndGC_DEBUG)
-            fprintf(stderr, "%s", os.str().c_str());
+        if (ndGC_DEBUG || ndGC.h_debug_log != stderr) {
+            fprintf(ndGC.h_debug_log, "%s", os.str().c_str());
+            if (ndGC.h_debug_log != stderr) fflush(ndGC.h_debug_log);
+        }
         os.str("");
     }
 
@@ -227,8 +216,10 @@ int ndDebugLogBufferUnlocked::sync() {
 
 int ndDebugLogBufferFlow::sync() {
     if (! os.str().empty()) {
-        if (ndGC_DEBUG || ndGC.h_flow != stderr)
-            fprintf(ndGC.h_flow, "%s", os.str().c_str());
+        if (ndGC_DEBUG || ndGC.h_debug_log_flow != stderr) {
+            fprintf(ndGC.h_debug_log_flow, "%s", os.str().c_str());
+            if (ndGC.h_debug_log_flow != stderr) fflush(ndGC.h_debug_log_flow);
+        }
         os.str("");
     }
 
@@ -523,7 +514,7 @@ void nd_private_ipaddr(uint8_t index, struct sockaddr_storage &addr) {
 
 bool nd_load_uuid(string &uuid, const string &path, size_t length) {
     struct stat sb;
-    char _uuid[length + 1];
+    unique_ptr<char[]> _uuid(new char[length + 1]);
 
     if (stat(path.c_str(), &sb) == -1) {
         if (errno != ENOENT) {
@@ -553,7 +544,7 @@ bool nd_load_uuid(string &uuid, const string &path, size_t length) {
 
         size_t bytes = 0;
 
-        bytes = fread((void *)_uuid, 1, length, ph);
+        bytes = fread((void *)_uuid.get(), 1, length, ph);
 
         int rc = pclose(ph);
 
@@ -578,7 +569,7 @@ bool nd_load_uuid(string &uuid, const string &path, size_t length) {
             return false;
         }
 
-        if (fread((void *)_uuid, 1, length, fh) != length) {
+        if (fread((void *)_uuid.get(), 1, length, fh) != length) {
             fclose(fh);
             nd_printf(
               "Error reading uuid from file: %s: %s\n",
@@ -591,7 +582,7 @@ bool nd_load_uuid(string &uuid, const string &path, size_t length) {
         _uuid[length] = '\0';
     }
 
-    uuid.assign(_uuid);
+    uuid.assign(_uuid.get());
 
     nd_rtrim(uuid);
 
@@ -688,8 +679,7 @@ const string &nd_get_version_and_features(bool fancy) {
         if (fancy) ident << ndTerm::Attr::BOLD;
         ident << ND_NAME << "/" << GIT_RELEASE;
         if (fancy) ident << ndTerm::Attr::RESET;
-        ident << " (" << os
-              << "; " << _ND_HOST_OS << "; " << _ND_HOST_CPU;
+        ident << " (" << os << "; " << _ND_HOST_OS << "; " << _ND_HOST_CPU;
 
         if (ndGC_USE_CONNTRACK) ident << "; conntrack";
         if (ndGC_USE_NETLINK) ident << "; netlink";
@@ -703,8 +693,6 @@ const string &nd_get_version_and_features(bool fancy) {
 #ifdef _ND_ENABLE_LIBTCMALLOC
         ident << "; tcmalloc";
 #endif
-        if (ndGC_SSL_USE_TLSv1) ident << "; ssl-tlsv1";
-        if (! ndGC_SSL_VERIFY) ident << "; ssl-no-verify";
 #ifdef HAVE_WORKING_REGEX
         ident << "; regex";
 #endif
@@ -800,8 +788,8 @@ int nd_file_load(const string &filename, string &data) {
 void nd_file_save(const string &filename, const string &data,
   bool append, mode_t mode, const char *user, const char *group) {
     int fd = open(filename.c_str(), O_WRONLY);
-    struct passwd *owner_user = nullptr;
-    struct group *owner_group = nullptr;
+    uid_t owner_user = -1;
+    gid_t owner_group = -1;
 
     if (fd < 0) {
         if (errno != ENOENT) {
@@ -814,26 +802,13 @@ void nd_file_save(const string &filename, const string &data,
               filename.c_str(), strerror(errno));
         }
 
-        if (user != nullptr) {
-            owner_user = getpwnam(user);
-            if (owner_user == nullptr) {
-                throw ndException("%s: getpwnam(%s): %s",
-                  __PRETTY_FUNCTION__, user, strerror(errno));
-            }
-        }
+        if (user != nullptr)
+            owner_user = nd_get_user_id(user);
 
-        if (group != nullptr) {
-            owner_group = getgrnam(group);
-            if (owner_group == nullptr) {
-                throw ndException("%s: getgrnam(%s): %s",
-                  __PRETTY_FUNCTION__, group, strerror(errno));
-            }
-        }
+        if (group != nullptr)
+            owner_group = nd_get_group_id(group);
 
-        if (fchown(fd,
-              (owner_user != nullptr) ? owner_user->pw_uid : -1,
-              (owner_group != nullptr) ? owner_group->gr_gid : -1) < 0)
-        {
+        if (fchown(fd, owner_user, owner_group) < 0) {
             throw ndException("%s: fchown(%s, %s, %s): %s",
               __PRETTY_FUNCTION__, (user != nullptr) ? user : "-1",
               (group != nullptr) ? group : "-1",
@@ -898,6 +873,11 @@ void nd_basename(const string &path, string &base) {
     base = path.substr(p + 1);
 }
 
+void nd_pathname(const string &path, string &base) {
+    size_t p = path.find_last_of("/");
+    if (p != string::npos) base = path.substr(0, p);
+}
+
 #if defined(__linux__)
 pid_t nd_is_running(pid_t pid, const string &exe_base) {
     pid_t rc = -1;
@@ -1532,6 +1512,30 @@ void ndTimer::Set(const struct itimerspec &itspec) {
     }
 }
 
+ndCPUSet::ndCPUSet(bool empty) {
+    if (empty) return;
+
+    cpu_set_t mask;
+    if (sched_getaffinity(0, sizeof(mask), &mask) == -1) {
+        throw ndExceptionSystemError(__PRETTY_FUNCTION__,
+          "sched_getaffinity");
+    }
+
+    for (int i = 0; i < CPU_SETSIZE; i++) {
+        if (! CPU_ISSET(i, &mask)) continue;
+        cpu_set.push_back(i);
+    }
+}
+
+ndCPUSet::ndCPUSet(const std::string &cpu_list) {
+}
+
+void ndCPUSet::Get(cpu_set_t &cpu_set) {
+    CPU_ZERO(&cpu_set);
+    for (auto& cpu : this->cpu_set)
+        CPU_SET(cpu, &cpu_set);
+}
+
 void nd_get_ip_protocol_name(int protocol, string &result) {
     static mutex lock;
     static unordered_map<int, string> cache;
@@ -1701,3 +1705,457 @@ void nd_time_ago(time_t seconds, string &ago) {
     ago = to_string((time_t)value) + " " + unit +
       ((plural) ? "s" : "");
 }
+
+std::ostream &operator<<(std::ostream &stream, ndApp::Id id) {
+    stream << static_cast<unsigned>(id);
+    return stream;
+}
+
+#if defined(__linux__)
+int nd_set_netns(const string &ns) {
+    string path = "/var/run/netns/" + ns;
+
+    int fd = open(path.c_str(), O_RDONLY);
+    if (fd < 0)
+        fd = open(ns.c_str(), O_RDONLY);
+    if (fd < 0) {
+        path = "/proc/" + ns + "/ns/net";
+        fd = open(path.c_str(), O_RDONLY);
+    }
+    if (fd < 0)
+        throw ndException("open: %s", strerror(errno));
+
+    int rc = setns(fd, CLONE_NEWNET);
+
+    close(fd);
+
+    if (rc < 0)
+        throw ndException("setns: %s", strerror(errno));
+
+    return rc;
+}
+
+void nd_get_memusage(const string &process, size_t &vm_kb, size_t &rss_kb) {
+    string pid, comm, state, ppid, pgrp, session, tty_nr;
+    string tpgid, flags, minflt, cminflt, majflt, cmajflt;
+    string utime, stime, cutime, cstime, priority, nice;
+    string O, itrealvalue, starttime;
+    size_t vsize, rss;
+
+    ifstream ifs("/proc/" + process + "/stat", ios_base::in);
+    ifs >> pid >> comm >> state >> ppid >> pgrp >> session >> tty_nr
+        >> tpgid >> flags >> minflt >> cminflt >> majflt >> cmajflt
+        >> utime >> stime >> cutime >> cstime >> priority >> nice
+        >> O >> itrealvalue >> starttime >> vsize >> rss;
+
+    static long page_size_kb = sysconf(_SC_PAGE_SIZE) / 1024;
+
+    vm_kb = vsize / 1024;
+    rss_kb = rss * page_size_kb;
+}
+#endif // __linux__
+
+uid_t nd_get_user_id(const string &name) {
+    const long bufsize = sysconf(_SC_GETPW_R_SIZE_MAX) > 0 ?
+        sysconf(_SC_GETPW_R_SIZE_MAX) : 16384;
+
+    unique_ptr<char[]> buffer(new char[bufsize]);
+
+    struct passwd pwd;
+    struct passwd *result;
+    int rc = getpwnam_r(name.c_str(),
+        &pwd, buffer.get(), bufsize, &result);
+
+    if (result == nullptr) {
+        if (rc == 0) {
+            throw ndException("%s: getpwnam_r(%s): %s",
+                __PRETTY_FUNCTION__, name.c_str(), strerror(ENOENT));
+        }
+
+        errno = rc;
+        throw ndException("%s: getpwnam_r(%s): %s",
+            __PRETTY_FUNCTION__, name.c_str(), strerror(errno));
+    }
+
+    return pwd.pw_uid;
+}
+
+gid_t nd_get_group_id(const string &name) {
+    const long bufsize = sysconf(_SC_GETGR_R_SIZE_MAX) > 0 ?
+        sysconf(_SC_GETGR_R_SIZE_MAX) : 16384;
+
+    unique_ptr<char[]> buffer(new char[bufsize]);
+
+    struct group grp;
+    struct group *result;
+    int rc = getgrnam_r(name.c_str(),
+        &grp, buffer.get(), bufsize, &result);
+
+    if (result == nullptr) {
+        if (rc == 0) {
+            throw ndException("%s: getgrnam_r(%s): %s",
+                __PRETTY_FUNCTION__, name.c_str(), strerror(ENOENT));
+        }
+
+        errno = rc;
+        throw ndException("%s: getgrnam_r(%s): %s",
+            __PRETTY_FUNCTION__, name.c_str(), strerror(errno));
+    }
+
+    return grp.gr_gid;
+}
+
+void nd_drop_privileges(const std::string &user, const std::string &group) {
+#ifdef _ND_ENABLE_LIBCAP
+    cap_t ctx = cap_get_proc();
+
+    const vector<cap_value_t> caps = { CAP_NET_ADMIN, CAP_NET_RAW };
+
+    cap_set_flag(ctx, CAP_PERMITTED, caps.size(), caps.data(), CAP_SET);
+
+    if (cap_set_proc(ctx)) {
+        throw ndException("%s: cap_set_proc(): %s",
+          __PRETTY_FUNCTION__, strerror(errno));
+    }
+
+    if (prctl(PR_SET_KEEPCAPS, 1L)) {
+        throw ndException("%s: prctl(PR_SET_KEEPCAPS): %s",
+          __PRETTY_FUNCTION__, strerror(errno));
+    }
+
+    cap_free(ctx);
+#endif // _ND_ENABLE_LIBCAP
+
+    uid_t uid = -1;
+    gid_t gid = -1;
+    if (! group.empty()) gid = nd_get_group_id(group);
+
+    if (! user.empty()) {
+        uid = nd_get_user_id(user);
+        initgroups(user.c_str(), gid);
+    }
+
+    if (setgid(gid) > 0) {
+        throw ndException("%s: setgid(%d): %s",
+          __PRETTY_FUNCTION__, gid, strerror(errno));
+    }
+
+    if (setuid(uid) > 0) {
+        throw ndException("%s: setuid(%d): %s",
+          __PRETTY_FUNCTION__, uid, strerror(errno));
+    }
+
+#ifdef _ND_ENABLE_LIBCAP
+    ctx = cap_get_proc();
+    cap_set_flag(ctx, CAP_EFFECTIVE, caps.size(), caps.data(), CAP_SET);
+
+    if (cap_set_proc(ctx)) {
+        throw ndException("%s: cap_set_proc(): %s",
+          __PRETTY_FUNCTION__, strerror(errno));
+    }
+
+    cap_free(ctx);
+#endif //
+}
+
+void nd_enable_coredumps(bool enable) {
+    struct rlimit limit = {
+        .rlim_cur = (enable) ? RLIM_INFINITY : 0,
+        .rlim_max = (enable) ? RLIM_INFINITY : 0
+    };
+
+    if (setrlimit(RLIMIT_CORE, &limit) < 0) {
+        nd_printf("WARNING: Unable to %s coredumps.\n",
+            (enable) ? "enable" : "disable", strerror(errno));
+    }
+}
+
+void ndTimeOfDay::Load(ndTimeOfDay &tod, const json &jconfig) {
+    tod.valid = false;
+
+    // Optional: day, default: all days
+    tod.days = ndTod_to_days(Day::NONE);
+
+    auto jdays = jconfig.find("day");
+    if (jdays != jconfig.end()) {
+        if (! jdays->is_array())
+            throw ndException("day must be an array");
+
+        for (auto &jd : *jdays) {
+            if (! jd.is_string()) continue;
+
+            string day = jd.get<string>().c_str();
+            const char *d = day.c_str();
+
+            if (strncasecmp("all", d, 3) == 0) {
+                tod.days = ndTod_to_days(Day::ALL);
+                break;
+            }
+
+            if (strncasecmp("sun", d, 3) == 0)
+                tod.days |= ndTod_to_days(Day::SUN);
+            else if (strncasecmp("mon", d, 3) == 0)
+                tod.days |= ndTod_to_days(Day::MON);
+            else if (strncasecmp("tue", d, 3) == 0)
+                tod.days |= ndTod_to_days(Day::TUE);
+            else if (strncasecmp("wed", d, 3) == 0)
+                tod.days |= ndTod_to_days(Day::WED);
+            else if (strncasecmp("thu", d, 3) == 0)
+                tod.days |= ndTod_to_days(Day::THU);
+            else if (strncasecmp("fri", d, 3) == 0)
+                tod.days |= ndTod_to_days(Day::FRI);
+            else if (strncasecmp("sat", d, 3) == 0)
+                tod.days |= ndTod_to_days(Day::SAT);
+            else
+                throw ndException("invalid day of the week: %s", d);
+        }
+    }
+    else tod.days = ndTod_to_days(Day::ALL);
+
+    // Mandatory: time_start
+    auto jtime_start = jconfig.find("time_start");
+    if (jtime_start == jconfig.end())
+        throw ndException("time_start not found");
+    if (! jtime_start->is_string())
+        throw ndException("time_start must be a string");
+    tod.time_start = ParseTime(jtime_start->get<string>());
+
+    // Mandatory: time_end
+    auto jtime_end = jconfig.find("time_end");
+    if (jtime_end == jconfig.end())
+        throw ndException("time_end not found");
+    if (! jtime_end->is_string())
+        throw ndException("time_end must be a string");
+    tod.time_end = ParseTime(jtime_end->get<string>());
+
+    // Optional: policy, default: inclusive
+    auto jpolicy = jconfig.find("policy");
+    if (jpolicy != jconfig.end()) {
+        if (! jpolicy->is_string())
+            throw ndException("day must be an string: inclusive, or exclusive");
+
+        string policy = jpolicy->get<string>();
+        if (strncasecmp("inc", policy.c_str(), 3) == 0)
+            tod.policy = Policy::INCLUDE;
+        else if (strncasecmp("exc", policy.c_str(), 3) == 0)
+            tod.policy = Policy::EXCLUDE;
+        else
+            throw ndException("invalid policy: %s", policy.c_str());
+    }
+    else tod.policy = Policy::INCLUDE;
+
+    tod.valid = true;
+}
+
+void ndTimeOfDay::Load(ndTimeOfDay &tod,
+    const string &time_start, const string &time_end,
+    ndTimeOfDay::Days days, ndTimeOfDay::Policy policy) {
+
+    tod.valid = false;
+
+    tod.time_start = ParseTime(time_start);
+    tod.time_end = ParseTime(time_end);
+    tod.days = days;
+    tod.policy = policy;
+
+    if (time_start > time_end)
+        throw ndException("%s: time_start > time_end", __PRETTY_FUNCTION__);
+
+    if (time_start == time_end)
+        throw ndException("%s: time_start == time_end", __PRETTY_FUNCTION__);
+
+    tod.valid = true;
+}
+
+time_t ndTimeOfDay::ParseTime(const string &time_string) {
+    size_t p = time_string.find_first_of(":");
+    if (p == string::npos)
+        throw ndException("%s: invalid time_string", __PRETTY_FUNCTION__);
+
+    time_t time_hour = (time_t)strtold(
+        time_string.substr(0, p).c_str(), nullptr);
+
+    if (time_hour > 23)
+        throw ndException("%s: invalid hour (> 23)", __PRETTY_FUNCTION__);
+
+    time_t time_min = (time_t)strtold(
+        time_string.substr(p + 1).c_str(), nullptr);
+    if (time_hour > 59)
+        throw ndException("%s: invalid minute (> 59)", __PRETTY_FUNCTION__);
+
+    return (time_hour * 3600) + (time_min * 60);
+}
+
+void ndTimeOfDay::Describe(const ndTimeOfDay &tod, string &desc) {
+    ostringstream os;
+
+    if (! tod.valid) os << "[invalid] ";
+
+    os << ((tod.policy == ndTimeOfDay::Policy::INCLUDE) ?
+            "within " : "outside of ");
+    os << tod.time_start << "s and " << tod.time_end << "s, on:";
+
+    if (tod.days == ndTod_to_days(Day::NONE)) os << " no days";
+    else if (tod.days == ndTod_to_days(Day::ALL)) {
+        os << " all days";
+    }
+    else {
+        if (tod.days & ndTod_to_days(Day::SUN)) os << " SUN";
+        if (tod.days & ndTod_to_days(Day::MON)) os << " MON";
+        if (tod.days & ndTod_to_days(Day::TUE)) os << " TUE";
+        if (tod.days & ndTod_to_days(Day::WED)) os << " WED";
+        if (tod.days & ndTod_to_days(Day::THU)) os << " THU";
+        if (tod.days & ndTod_to_days(Day::FRI)) os << " FRI";
+        if (tod.days & ndTod_to_days(Day::SAT)) os << " SAT";
+    }
+
+    desc = os.str();
+}
+
+void ndTimeOfDay::UpdateTimezone(void) {
+    try {
+        if (timezone.empty()) {
+            string tz_error;
+
+            try {
+                civil_time::timezone tz;
+
+                timezone = tz.name();
+                timezone_abbrv = tz.abbrv();
+                is_dst = tz.is_dst();
+                utc_offset = tz.utc_offset();
+
+            } catch (runtime_error const &e) {
+                tz_error = "runtime_error: localtime: ";
+                tz_error.append(e.what());
+            } catch (invalid_argument const &e) {
+                tz_error = "invalid_argument: localtime: ";
+                tz_error.append(e.what());
+            }
+
+            if (! tz_error.empty()) {
+
+                nd_dprintf(
+                    "WARNING: unable to set timezone: %s, trying UTC...\n",
+                    tz_error.c_str());
+
+                civil_time::timezone tz("UTC");
+
+                timezone = tz.name();
+                timezone_abbrv = tz.abbrv();
+                is_dst = tz.is_dst();
+                utc_offset = tz.utc_offset();
+            }
+        }
+        else {
+            civil_time::timezone tz(timezone);
+
+            timezone = tz.name();
+            timezone_abbrv = tz.abbrv();
+            is_dst = tz.is_dst();
+            utc_offset = tz.utc_offset();
+        }
+    } catch (runtime_error const &e) {
+        ndException("runtime_error: %s: %s",
+            timezone.empty() ? "UTC" : timezone.c_str(),
+            e.what());
+    } catch (invalid_argument const &e) {
+        throw ndException("invalid_argument: %s: %s",
+            timezone.empty() ? "UTC" : timezone.c_str(),
+            e.what());
+    }
+
+    nd_dprintf("%s: %s (%s): utc_offset: %d, is_dst: %s\n",
+        __PRETTY_FUNCTION__,
+        timezone.empty() ? "local" : timezone.c_str(),
+        timezone_abbrv.empty() ? "???" : timezone_abbrv.c_str(),
+        utc_offset, (is_dst) ? "yes" : "no");
+}
+
+time_t ndTimeOfDay::TickUpdate(void) {
+    time_t tv_now = nd_time_monotonic();
+
+    if (tv_last_update + tv_update_freq < tv_now) {
+            UpdateTimezone();
+            tv_last_update = tv_now;
+    }
+
+    return time(nullptr) + utc_offset;
+}
+
+bool ndTimeOfDay::operator==(const time_t tv) const {
+    if (! valid) return false;
+    if (days == ndTod_to_days(Day::NONE)) return false;
+
+    struct tm tm = { 0 };
+
+    if (gmtime_r(&tv, &tm) != &tm) {
+        throw ndException("%s: %s: %s", __PRETTY_FUNCTION__,
+          "gmtime_r", strerror(errno));
+    }
+
+    if (days != ndTod_to_days(Day::ALL)) {
+        switch (tm.tm_wday) {
+        case 0:
+            if (policy == Policy::INCLUDE && ! (days & ndTod_to_days(Day::SUN)))
+                return false;
+            if (policy == Policy::EXCLUDE && (days & ndTod_to_days(Day::SUN)))
+                return false;
+            break;
+        case 1:
+            if (policy == Policy::INCLUDE && ! (days & ndTod_to_days(Day::MON)))
+                return false;
+            if (policy == Policy::EXCLUDE && (days & ndTod_to_days(Day::MON)))
+                return false;
+            break;
+        case 2:
+            if (policy == Policy::INCLUDE && ! (days & ndTod_to_days(Day::TUE)))
+                return false;
+            if (policy == Policy::EXCLUDE && (days & ndTod_to_days(Day::TUE)))
+                return false;
+            break;
+        case 3:
+            if (policy == Policy::INCLUDE && ! (days & ndTod_to_days(Day::WED)))
+                return false;
+            if (policy == Policy::EXCLUDE && (days & ndTod_to_days(Day::WED)))
+                return false;
+            break;
+        case 4:
+            if (policy == Policy::INCLUDE && ! (days & ndTod_to_days(Day::THU)))
+                return false;
+            if (policy == Policy::EXCLUDE && (days & ndTod_to_days(Day::THU)))
+                return false;
+            break;
+        case 5:
+            if (policy == Policy::INCLUDE && ! (days & ndTod_to_days(Day::FRI)))
+                return false;
+            if (policy == Policy::EXCLUDE && (days & ndTod_to_days(Day::FRI)))
+                return false;
+            break;
+        case 6:
+            if (policy == Policy::INCLUDE && ! (days & ndTod_to_days(Day::SAT)))
+                return false;
+            if (policy == Policy::EXCLUDE && (days & ndTod_to_days(Day::SAT)))
+                return false;
+            break;
+        }
+    }
+
+    tm.tm_sec = tm.tm_min = tm.tm_hour = 0;
+
+    time_t tv_today = timegm(&tm);
+    if (tv_today == (time_t)-1) {
+        throw ndException(
+            "%s: timegm: %s", __PRETTY_FUNCTION__, strerror(errno));
+    }
+
+    time_t tv_seconds = tv - tv_today;
+
+    if (policy == Policy::INCLUDE &&
+        (tv_seconds < time_start || tv_seconds > time_end)) return false;
+
+    if (policy == Policy::EXCLUDE &&
+        tv_seconds >= time_start && tv_seconds <= time_end) return false;
+
+    return true;
+}
